# (C)2018 Ingram Micro Inc. all Rights Reserved
from sh_base import CleanUpSHBase

from oacleanup.settings import OAModule, CleanUpSteps
from oacleanup.utils import packages, hosts, resources
from oacleanup.utils.billing import is_billing_installed, get_billing_api
from oacleanup.utils.errors import CleanUpException

from poaupdater import uConst, uSysDB, openapi

from collections import defaultdict

import logging

logger = logging.getLogger()

EXCHANGE_SERVICES = [
    ('service', 'Exchange2010OAB'),
    ('service', 'Exchange2010IMAP4'),
    ('service', 'Exchange2010SMTPAuth'),
    ('service', 'Exchange2010Protocols'),
    ('service', 'Exchange2010SMTP'),
    ('service', 'Exchange2010POP3'),
    ('service', 'Exchange2010Autodiscover'),
    ('service', 'Exchange2010Mailstore'),

    ('service', 'Exchange2013OAB'),
    ('service', 'Exchange2013IMAP4'),
    ('service', 'Exchange2013SMTPAuth'),
    ('service', 'Exchange2013Protocols'),
    ('service', 'Exchange2013SMTP'),
    ('service', 'Exchange2013POP3'),
    ('service', 'Exchange2013Autodiscover'),
    ('service', 'Exchange2013Mailstore'),

    ('service', 'Exchange2016OAB'),
    ('service', 'Exchange2016IMAP4'),
    ('service', 'Exchange2016SMTPAuth'),
    ('service', 'Exchange2016Protocols'),
    ('service', 'Exchange2016SMTP'),
    ('service', 'Exchange2016POP3'),
    ('service', 'Exchange2016Autodiscover'),
    ('service', 'Exchange2016Mailstore'),
]

EXCHANGE_MN_PACKAGES = [
    ('other', 'PrivilegesExchange'),
    ('other', 'ExchangeApplication'),
]

EXCHANGE_SN_PACKAGES = EXCHANGE_SERVICES + [
    ('other', 'MPFProviderClients'),
    ('other', '.NET2'),
]

EXCHANGE_PACKAGES_TO_REMOVE_FROM_REPO = EXCHANGE_MN_PACKAGES + EXCHANGE_SN_PACKAGES + [
    ('other', 'ExchangeMigration')
]

EXCHANGE_SCS = [
    'Exchange'
]

EXCHANGE_MODULE_NAME = 'Messaging and Collaboration'

help_link = 'https://kb.cloudblue.com/en/133390'


class CleanUpHostedExchange(CleanUpSHBase):
    def __init__(self):
        list_of_available_steps = [
            CleanUpSteps.DEACTIVATE_SERVICE_TEMPLATES,
            CleanUpSteps.CLEAR_SERVICE_NODES,
            CleanUpSteps.UNINSTALL_MODULE,
        ]
        if is_billing_installed():
            list_of_available_steps.insert(1, CleanUpSteps.SYNCHRONIZE_ONLINE_STORES)

        super(CleanUpHostedExchange, self).__init__(
            name=OAModule.EXCHANGE,
            list_of_available_steps=list_of_available_steps,
            indicator_services=EXCHANGE_SERVICES,
            services_to_detach=EXCHANGE_SERVICES,
            module_services=EXCHANGE_SN_PACKAGES,
            help_link=help_link,
        )
        self.api = openapi.OpenAPI()
        if is_billing_installed():
            self.pba_api = get_billing_api()

    def _remove_packages_from_repo(self, sc_list):
        count_packages = packages.remove_packages_from_repo(
            EXCHANGE_PACKAGES_TO_REMOVE_FROM_REPO, [uConst.Const.getOsaWinPlatform()]
        )
        scs = [('sc', sc) for sc in sc_list]
        count_packages += packages.remove_packages_from_repo(scs, None)
        return count_packages

    def _uninstall_module(self):
        scs_to_rm = EXCHANGE_SCS
        self._remove_resource_types(scs_to_rm)
        self._remove_aps_resource_types()
        self._remove_resource_classes(scs_to_rm)

        self._remove_sc_packages_from_mn(scs_to_rm)
        self._drop_scs_tables_seqs()
        packages.remove_packages_on_hosts([1], EXCHANGE_MN_PACKAGES)

        self._unregister_module(EXCHANGE_MODULE_NAME)
        self._remove_packages(scs_to_rm)
        hosts.restart_pa_agent()

    def _drop_scs_tables_seqs(self):
        con = uSysDB.connect()
        cur = con.cursor()
        self._drop_seqs_deps(['exch_contacts_addresses_seq',
                              'exch_mailbox_stores_seq',
                              'exch_public_folders_seq',
                              'exch_spla_sku_seq',
                              'exch_disclaimers_seq',
                              'exch_mailgroups_addresses_seq',
                              'exch_public_stores_seq',
                              'exch_temp_seq',
                              'exch_mailboxes_addresses_seq',
                              'exch_nlb_clusters_seq',
                              'exch_spla_features_seq',
                              'exch_mailbox_plans_seq',
                              'exch_public_folders_addresses_seq',
                              'exch_spla_reports_seq',
                              ],
                             cur)
        con.commit()

    def _find_rt(self, resclass_name, name):
        logger.debug("Find resource type by name '%s' and resource class name '%s'" % (name, resclass_name))
        _resources = filter(lambda x: x['resource_type_name'] == name,
                            self.api.pem.getResourceTypesByClass(
                                resclass_name=resclass_name,
                                show_system_rt=False
                            ))
        if _resources:
            # noinspection PyTypeChecker
            rt_id = _resources[0]['resource_type_id']
            logger.info("Found existing resource type '%s' (class %s) with ID #%s" % (name, resclass_name, rt_id))
            return rt_id
        return None

    def _remove_aps_resource_types(self):
        logger.info('Removing %s APS resources types', self.name)
        rts = [
            self._find_rt('rc.saas.service.link', 'Hosted Exchange Application'),
            self._find_rt('rc.saas.service', 'Hosted Exchange Organization')
        ]
        included_resources = {}
        if is_billing_installed():
            included_resources = defaultdict(list)
            for res in self.pba_api.IncludedResourcesGetIncludedResourcesList():
                comp_id = res[0]
                inc_id = res[1]
                included_resources[inc_id].append(comp_id)

        for rt_id in rts:
            try:
                if not rt_id:
                    continue
                service_templates_by_resource = resources.get_service_templates_by_resource(rt_id)

                for st_id in service_templates_by_resource:
                    logger.debug('Removing the resource type with ID %s from the service template with ID %s...' % (
                        rt_id, st_id
                    ))
                    if not resources.is_resource_in_service_template(st_id, rt_id):
                        logger.debug(
                            'The resource type with ID %s was already removed from the service template with ID %s' % (
                                rt_id, st_id
                            )
                        )
                        continue

                    self.api.pem.removeResourceTypeFromServiceTemplate(owner_id=1, st_id=st_id, rt_id=rt_id)
                    logger.debug(
                        'The resource type with ID %s was removed from the service template with ID %s' % (rt_id, st_id)
                    )

                if rt_id in included_resources:
                    for comp_id in included_resources[rt_id]:
                        self.pba_api.IncludedResourcesMoveToArc(comp_id, rt_id)

                logger.debug('Removing the resource type with ID %s', rt_id)
                resources.remove_resource_type(rt_id)
            except Exception as e:
                raise CleanUpException(
                    'Unable to remove the resource type #%d from service templates: %s. '
                    'Try to remove it from the Provider Panel.' % (rt_id, e)
                )
