# (C)2018 Ingram Micro Inc. all Rights Reserve
from poaupdater import openapi, uResourceManagement, uSysDB

from oacleanup.utils.billing import is_billing_installed, get_billing_api
from oacleanup.utils.errors import CleanUpException

from collections import defaultdict

import logging

logger = logging.getLogger()


def get_service_templates_by_resource(rt_id):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT st_id FROM service_templates'
             ' JOIN st_resources USING (st_id)'
             ' WHERE rt_id = %s')
    cur.execute(query, rt_id)

    return [st_id for st_id, in cur.fetchall()]


def get_resource_types_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rt_id FROM resource_types'
             ' JOIN resource_classes USING (class_id)'
             ' JOIN service_classes USING (sc_id)'
             ' WHERE service_classes.name = %s')
    cur.execute(query, sc_name)

    return [rt_id for rt_id, in cur.fetchall()]


def get_root_resource_types_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rt_id FROM resource_types'
             ' JOIN resource_classes USING (class_id)'
             ' JOIN service_classes USING (sc_id)'
             ' WHERE parent_class_id IS NULL'
             ' AND service_classes.name = %s')
    cur.execute(query, sc_name)

    return [rt_id for rt_id, in cur.fetchall()]


def get_nonroot_resource_classes_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rc.name FROM resource_classes rc'
             ' JOIN service_classes sc USING (sc_id)'
             ' WHERE parent_class_id IS NOT NULL'
             ' AND sc.name = %s'
             ' ORDER BY class_id DESC')
    cur.execute(query, sc_name)

    return [name for name, in cur.fetchall()]


def get_root_resource_classes_for_sc(sc_name):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ('SELECT rc.name FROM resource_classes rc'
             ' JOIN service_classes sc USING (sc_id)'
             ' WHERE parent_class_id IS NULL'
             ' AND sc.name = %s'
             ' ORDER BY class_id DESC')
    cur.execute(query, sc_name)

    return [name for name, in cur.fetchall()]


def remove_resource_type(rt_id):
    con = uSysDB.connect()
    cur = con.cursor()

    query = ("UPDATE resource_types rt SET system = 'n'"
             " WHERE rt_id = %s")

    cur.execute(query, rt_id)
    con.commit()

    api = openapi.OpenAPI()
    api.pem.removeResourceType(rt_id=rt_id)


def is_resource_in_service_template(st_id, rt_id):
    con = uSysDB.connect()
    cur = con.cursor()

    cur.execute(
        'SELECT st_id FROM service_templates JOIN st_resources USING (st_id) WHERE rt_id = %s AND st_id = %s',
        rt_id, st_id
    )

    return [st_id for st_id, in cur.fetchall()]


def remove_resource_types(service_controller_list):
    pba_api = None
    included_resources = {}
    if is_billing_installed():
        pba_api = get_billing_api()

        included_resources = defaultdict(list)
        for res in pba_api.IncludedResourcesGetIncludedResourcesList():
            comp_id = res[0]
            inc_id = res[1]
            included_resources[inc_id].append(comp_id)

    resource_count = 0
    api = openapi.OpenAPI()
    for sc_name in service_controller_list:
        for rt_id in get_resource_types_for_sc(sc_name):
            try:
                service_templates_by_resource = get_service_templates_by_resource(rt_id)
                if service_templates_by_resource:
                    resource_count += 1

                for st_id in service_templates_by_resource:
                    logger.debug('Removing the resource type with ID %s from the service template with ID %s...' % (
                        rt_id, st_id
                    ))
                    if not is_resource_in_service_template(st_id, rt_id):
                        logger.debug(
                            'The resource type with ID %s was already removed from the service template with ID %s' % (
                                rt_id, st_id
                            )
                        )
                        continue

                    api.pem.removeResourceTypeFromServiceTemplate(owner_id=1, st_id=st_id, rt_id=rt_id)
                    logger.debug(
                        'The resource type with ID %s was removed from the service template with ID %s' % (rt_id, st_id)
                    )

                if rt_id in included_resources:
                    for comp_id in included_resources[rt_id]:
                        pba_api.IncludedResourcesMoveToArc(comp_id, rt_id)
            except Exception as e:
                raise CleanUpException(
                    'Unable to remove the resource type #%d from service templates: %s. '
                    'Try to remove it from the Provider Panel.' % (rt_id, e)
                )

        for rt_id in get_root_resource_types_for_sc(sc_name):
            logger.debug('Removing the root resource type with ID %s', rt_id)
            remove_resource_type(rt_id)

    return resource_count


def remove_resource_classes(scs):
    rc_count = 0

    con = uSysDB.connect()
    api = openapi.OpenAPI()

    for sc_name in scs:
        for name in get_nonroot_resource_classes_for_sc(sc_name):
            logger.debug('Removing the non root resource class %s', name)
            uResourceManagement.deleteResourceClass(con, name)
            rc_count += 1
        con.commit()

        for name in get_root_resource_classes_for_sc(sc_name):
            logger.debug('Removing the root resource class %s', name)
            api.pem.unregisterResourceClass(name=name)
            rc_count += 1

    return rc_count


def get_resource_types_for_services_by_subscriptions(service_names, subscription_ids):
    if not service_names:
        return []

    # resource having at least one attribute
    q = ('SELECT DISTINCT sub_id, rt_id, restype_name, rt_attributes.attribute'
         ' FROM subs_resources'
         ' JOIN resource_types USING (rt_id)'
         ' JOIN resource_classes USING (class_id)'
         ' JOIN services USING (sc_id)'
         ' JOIN rt_attributes USING (rt_id)'
         ' JOIN host_attributes USING (host_id)'
         ' WHERE rt_attributes.attribute = host_attributes.host_attribute'
         ' AND services.name IN %s'
         ' AND sub_limit != 0'
         ' AND subs_resources.sub_id IN %s')

    con = uSysDB.connect()
    cur = con.cursor()

    cur.execute(q, (tuple(service_names), tuple(subscription_ids)))
    rt_with_attrs = [
        (sub_id, rt_id, restype_name, rt_attr) for [sub_id, rt_id, restype_name, rt_attr] in cur.fetchall()
        ]

    # resource w/o attribute
    q = ('SELECT DISTINCT sub_id, rt_id, restype_name'
         ' FROM subs_resources'
         ' JOIN resource_types USING (rt_id)'
         ' JOIN resource_classes USING (class_id)'
         ' JOIN services USING (sc_id)'
         ' LEFT JOIN rt_attributes USING (rt_id)'
         ' WHERE services.name IN %s'
         ' AND sub_limit != 0'
         ' AND subs_resources.sub_id IN %s'
         ' GROUP BY sub_id, rt_id, restype_name'
         ' HAVING COUNT(rt_attributes.rt_id) = 0')

    cur.execute(q, (tuple(service_names), tuple(subscription_ids)))
    rt_no_attrs = [(sub_id, rt_id, restype_name, None) for [sub_id, rt_id, restype_name] in cur.fetchall()]

    return set(rt_with_attrs + rt_no_attrs)
